@@ -2,14 +2,19 @@ import argparse, time, re, asyncio, functools, base64, random, urllib.parse, soc
 from . import proto
 from .__doc__ import *
 
-SOCKET_TIMEOUT = 300
+SOCKET_TIMEOUT = 60
 UDP_LIMIT = 30
 DUMMY = lambda s: s
 
-asyncio.StreamReader.read_w = lambda self, n: asyncio.wait_for(self.read(n), timeout=SOCKET_TIMEOUT)
-asyncio.StreamReader.read_n = lambda self, n: asyncio.wait_for(self.readexactly(n), timeout=SOCKET_TIMEOUT)
-asyncio.StreamReader.read_until = lambda self, s: asyncio.wait_for(self.readuntil(s), timeout=SOCKET_TIMEOUT)
-asyncio.StreamReader.rollback = lambda self, s: self._buffer.__setitem__(slice(0, 0), s)
+def patch_StreamReader(c=asyncio.StreamReader):
+    c.read_w = lambda self, n: asyncio.wait_for(self.read(n), timeout=SOCKET_TIMEOUT)
+    c.read_n = lambda self, n: asyncio.wait_for(self.readexactly(n), timeout=SOCKET_TIMEOUT)
+    c.read_until = lambda self, s: asyncio.wait_for(self.readuntil(s), timeout=SOCKET_TIMEOUT)
+    c.rollback = lambda self, s: self._buffer.__setitem__(slice(0, 0), s)
+def patch_StreamWriter(c=asyncio.StreamWriter):
+    c.is_closing = lambda self: self._transport.is_closing() # Python 3.6 fix
+patch_StreamReader()
+patch_StreamWriter()
 
 class AuthTable(object):
     _auth = {}
@@ -220,9 +225,7 @@ class ProxyDirect(object):
         except Exception as ex:
             raise
         return reader, writer
-    def prepare_connection(self, reader_remote, writer_remote, host, port):
-        return self.prepare_ciphers_and_headers(reader_remote, writer_remote, host, port)
-    async def prepare_ciphers_and_headers(self, reader_remote, writer_remote, host, port):
+    async def prepare_connection(self, reader_remote, writer_remote, host, port):
         return reader_remote, writer_remote
     async def tcp_connect(self, host, port, local_addr=None, lbind=None):
         reader, writer = await self.open_connection(host, port, local_addr, lbind)
@@ -288,12 +291,12 @@ class ProxySimple(ProxyDirect):
             return asyncio.open_unix_connection(path=self.bind)
         else:
             return asyncio.open_connection(host=self.host_name, port=self.port, local_addr=local_addr, family=family)
-    async def prepare_ciphers_and_headers(self, reader_remote, writer_remote, host, port):
+    async def prepare_connection(self, reader_remote, writer_remote, host, port):
         reader_remote, writer_remote = proto.sslwrap(reader_remote, writer_remote, self.sslclient, False, self.host_name)
         _, writer_cipher_r = await prepare_ciphers(self.cipher, reader_remote, writer_remote, self.bind)
         whost, wport = self.jump.destination(host, port)
         await self.rproto.connect(reader_remote=reader_remote, writer_remote=writer_remote, rauth=self.auth, host_name=whost, port=wport, writer_cipher_r=writer_cipher_r, myhost=self.host_name, sock=writer_remote.get_extra_info('socket'))
-        return await self.jump.prepare_ciphers_and_headers(reader_remote, writer_remote, host, port)
+        return await self.jump.prepare_connection(reader_remote, writer_remote, host, port)
     def start_server(self, args, stream_handler=stream_handler):
         handler = functools.partial(stream_handler, **vars(self), **args)
         if self.unix:
@@ -315,9 +318,7 @@ class ProxyQUIC(ProxySimple):
         writer.get_extra_info = dict(peername=remote_addr, sockname=remote_addr).get
         writer.drain = drain
         closed = False
-        def is_closing():
-            return closed
-        writer.is_closing = is_closing
+        writer.is_closing = lambda: closed
         def close():
             nonlocal closed
             closed = True
@@ -326,7 +327,7 @@ class ProxyQUIC(ProxySimple):
             except Exception:
                 pass
         writer.close = close
-    async def wait_open_connection(self, *args):
+    async def wait_quic_connection(self):
         if self.handshake is not None:
             if not self.handshake.done():
                 await self.handshake
@@ -340,27 +341,53 @@ class ProxyQUIC(ProxySimple):
                     elif isinstance(event, aioquic.quic.events.ConnectionTerminated):
                         self.handshake = None
                         self.quic_egress_acm = None
+                    elif isinstance(event, aioquic.quic.events.StreamDataReceived):
+                        if event.stream_id in self.udpmap:
+                            self.udpmap[event.stream_id](self.udp_packet_unpack(event.data))
+                            return
                     super().quic_event_received(event)
             self.quic_egress_acm = aioquic.asyncio.connect(self.host_name, self.port, create_protocol=Protocol, configuration=self.quicclient)
             conn = await self.quic_egress_acm.__aenter__()
             await self.handshake
+    async def udp_open_connection(self, host, port, data, addr, reply):
+        await self.wait_quic_connection()
+        conn = self.handshake.result()
+        if addr in self.udpmap:
+            stream_id = self.udpmap[addr]
+        else:
+            stream_id = conn._quic.get_next_available_stream_id(False)
+            self.udpmap[addr] = stream_id
+            self.udpmap[stream_id] = reply
+            conn._quic._get_or_create_stream_for_send(stream_id)
+        conn._quic.send_stream_data(stream_id, data, False)
+        conn.transmit()
+    async def wait_open_connection(self, *args):
+        await self.wait_quic_connection()
         conn = self.handshake.result()
         stream_id = conn._quic.get_next_available_stream_id(False)
         conn._quic._get_or_create_stream_for_send(stream_id)
         reader, writer = conn._create_stream(stream_id)
         self.patch_writer(writer)
         return reader, writer
+    async def udp_start_server(self, args):
+        import aioquic.asyncio, aioquic.quic.events
+        class Protocol(aioquic.asyncio.QuicConnectionProtocol):
+            def quic_event_received(s, event):
+                if isinstance(event, aioquic.quic.events.StreamDataReceived):
+                    stream_id = event.stream_id
+                    addr = ('quic '+self.bind, stream_id)
+                    event.sendto = lambda data, addr: (s._quic.send_stream_data(stream_id, data, False), s.transmit())
+                    event.get_extra_info = {}.get
+                    asyncio.ensure_future(datagram_handler(event, event.data, addr, **vars(self), **args))
+                    return
+                super().quic_event_received(event)
+        return await aioquic.asyncio.serve(self.host_name, self.port, configuration=self.quicserver, create_protocol=Protocol), None
     def start_server(self, args, stream_handler=stream_handler):
         import aioquic.asyncio
         def handler(reader, writer):
             self.patch_writer(writer)
             asyncio.ensure_future(stream_handler(reader, writer, **vars(self), **args))
-        return aioquic.asyncio.serve(
-            self.host_name,
-            self.port,
-            configuration=self.quicserver,
-            stream_handler=handler
-        )
+        return aioquic.asyncio.serve(self.host_name, self.port, configuration=self.quicserver, stream_handler=handler)
 
 class ProxySSH(ProxySimple):
     def __init__(self, **kw):
@@ -371,7 +398,7 @@ class ProxySSH(ProxySimple):
     def patch_stream(self, ssh_reader, writer, host, port):
         reader = asyncio.StreamReader()
         async def channel():
-            while not writer.is_closing():
+            while not ssh_reader.at_eof() and not writer.is_closing():
                 buf = await ssh_reader.read(65536)
                 if not buf:
                     break
@@ -413,6 +440,8 @@ class ProxySSH(ProxySimple):
             reader, writer = self.patch_stream(reader, writer, host, port)
         return reader, writer
     async def start_server(self, args, stream_handler=stream_handler, tunnel=None):
+        if type(self.jump) is ProxyDirect:
+            raise Exception('ssh server mode unsupported')
         await self.wait_ssh_connection(tunnel=tunnel)
         conn = self.sshconn.result()
         if isinstance(self.jump, ProxySSH):
@@ -442,7 +471,7 @@ class ProxyBackward(ProxySimple):
     async def wait_open_connection(self, *args):
         while True:
             reader, writer = await self.conn.get()
-            if not writer.is_closing():
+            if not reader.at_eof() and not writer.is_closing():
                 return reader, writer
     def close(self):
         self.closed = True
@@ -652,19 +681,19 @@ def main():
             if option.sslclient:
                 option.sslclient.load_cert_chain(*sslfile)
                 option.sslserver.load_cert_chain(*sslfile)
-        for option in args.listen+args.rserver:
+        for option in args.listen+args.ulisten+args.rserver+args.urserver:
             if isinstance(option, ProxyQUIC):
                 option.quicserver.load_cert_chain(*sslfile)
             if isinstance(option, ProxyBackward) and isinstance(option.backward, ProxyQUIC):
                 option.backward.quicserver.load_cert_chain(*sslfile)
-    elif any(map(lambda o: o.sslclient or isinstance(o, ProxyQUIC), args.listen)):
+    elif any(map(lambda o: o.sslclient or isinstance(o, ProxyQUIC), args.listen+args.ulisten)):
         print('You must specify --ssl to listen in ssl mode')
         return
     if args.test:
         asyncio.get_event_loop().run_until_complete(test_url(args.test, args.rserver))
         return
     if not args.listen and not args.ulisten:
-        args.listen.append(proxy_by_uri('http+socks4+socks5://:8080/'))
+        args.listen.append(proxies_by_uri('http+socks4+socks5://:8080/'))
     args.httpget = {}
     if args.pac:
         pactext = 'function FindProxyForURL(u,h){' + (f'var b=/^(:?{args.block.__self__.pattern})$/i;if(b.test(h))return "";' if args.block else '')
